# External libraries
from z3 import *             
import copy
import math
import pandas as pd

# Planet system imports
from planet.unit import Groups
from planet.variable import  multifact, ExperimentVariable, MultiFactVariable
from planet.design import Design
from planet.constraint import (
    StartWith, Counterbalance, NoRepeat, InnerBlock, OuterBlock,
    SetRank, SetPosition, AbsoluteRank
)
from planet.candl import combine_lists

def eval(designs):
    for design in designs:
        if design.num_plans() == 0:
            design._determine_num_plans()
        
def is_counterbalanced(d1, d2):
    return d1.counterbalanced or d2.counterbalanced

def nest_structure(d1, d2):
    constraints = []
    # Match all variables from the outer design within each block matrix
    
    for variable in d2.design_variables:
        constraints.append(InnerBlock(
            variable,
            d1.get_width(),
            d1.num_plans(),
            stride = [1, 1]
        ))

     # Match all variables from the inner design across every block

    for variable in d1.design_variables:
        constraints.append(OuterBlock(
            variable,
            d1.get_width(),
            d1.num_plans(),
            stride = [1, 1]
        ))


    return constraints


def copy_nested_constraints(design1, design2):
    width1, width2 = design1.get_width(), design2.get_width()
    total_groups = design1.num_plans() * design2.num_plans()
    total_conditions = width1 * width2
    constraints = []
    # need to modify the inner constraints region
    # Add counterbalance constraints from design1
    for constraint in design1.constraints.get_constraints():
        if isinstance(constraint, Counterbalance):
            constraints.append(
                Counterbalance(
                    constraint.variable,
                    width=constraint.width or design1.get_width(),
                    height=constraint.height or design1.num_plans(),
                    stride=constraint.stride
                )
            )

        elif isinstance(constraint, NoRepeat):
            constraints.append(
                NoRepeat(
                    constraint.variable,
                    width=constraint.width if constraint.width < design1.get_width() else design1.get_width(),
                    stride=constraint.stride
                )
            )

        else:
            constraints.append(copy.copy(constraint))


    # need to modify out constraint region
    for constraint in design2.constraints.get_constraints():
        # FIXME FIXME, I think this is a ratio problem
        if isinstance(constraint, Counterbalance):
            stride = [design1.num_plans(), width1 * constraint.stride[1]]
            # Add counterbalance constraint for design2 variables
            constraints.append(
                Counterbalance(
                    constraint.variable,
                    width=total_conditions,
                    height=total_groups,
                    stride=stride
                )
            )

        elif isinstance(constraint, NoRepeat):
            width = constraint.width if constraint.width < width2 else width2
            constraints.append(
                NoRepeat(
                    constraint.variable,
                    width*width1,
                    constraint.stride*width1
                )
            )
    
        elif isinstance(constraint, InnerBlock):
            constraints.append(
                InnerBlock(
                    constraint.variable, 
                    constraint.width*width1, 
                    constraint.height*design1.num_plans(), 
                    stride = [1, 1])
            )

        # # here I need to multiply stride by the number of conditions of the block variable
        elif isinstance(constraint, OuterBlock):
            constraints.append(
                OuterBlock(
                    constraint.variable, 
                    constraint.width*width1, 
                    constraint.height*design1.num_plans(), 
                    stride = [1, 1])
            )
            
        else:
            constraints.append(copy.copy(constraint))
            
    return constraints


def add_design_variables(*, des, combined_des):
    for v in des.design_variables:
        combined_des.add_variable(v)


def nest(*, outer:Design, inner:Design):
    """
    Nest two designs to create a combined experimental design.

    Args:
        outer: The outer design object.
        inner: The inner design object.

    Returns:
        A new Design object representing the nested experimental design.
    """
    total_groups = inner.num_plans() * outer.num_plans()
    total_conditions = inner.get_width() * outer.get_width()

    combined_design = (
        Design()
        .limit_groups(total_groups)
        .num_trials(total_conditions)
    )


    # NOTE: ORDER MATTERS HERE... The design variable spec stores one outermatch
    # constraint. Prioritze existing constraints, as they will always have a
    # width equal to or smaller. Make this more elegant later?
    # combined_design.add_variables(combined_variables)
    add_design_variables(des = inner, combined_des = combined_design)
    add_design_variables(des = outer, combined_des = combined_design)
    # Nest structural constraints and copy semantic ones
    copied_constraints = copy_nested_constraints(inner, outer)
    if copied_constraints:
        combined_design.add_constraints(copied_constraints)

    combined_design.add_constraints(nest_structure(inner, outer))


    return combined_design







# External dependencies
from z3 import *  # requires `pip install z3-solver`
import math
import pandas as pd
# Internal modules (installed via `pip install -e .`)
from planet.unit import Groups
from planet.variable import MultiFactVariable, multifact
from planet.constraint import (
    StartWith, Counterbalance, NoRepeat,
    InnerBlock, OuterBlock, Constraint,
    SetRank, SetPosition, AbsoluteRank
)
from planet.designer import Designer
from planet.randomizer import Randomizer
from planet.formatter import LatexExport
from planet.candl import *
from planet.candl import generate_conditions
from planet.helpers import *
from planet.narray import *
from planet.constraint_manager import ConstraintManager
import hashlib
from planet.design_variable import DesignVariable
from planet.design_exceptions import *

  
# # TODO: 
#     1. better handling of rand_variables
#     2. separate constraintspec from the actual constraint
#     3. identify patter when copying constraints (in nest and cross)

class Design:
    """Main class for creating experimental designs."""
    def __init__(self):
        self.variables = [] 
        self.groups = Groups()
        self.constraints = ConstraintManager()
        self.trials = 0
        self.designer = Designer()
        self.previous_snapshot = None
        self.design_variables = {}

    @property
    def is_constrained(self):
        # FIXME: not handled for replicated trials
        return self.constraints.check_property(lambda c: isinstance(c, (OuterBlock, InnerBlock)))
    
    @property
    def counterbalanced(self):
        # FIXME: not handled for replicated trials
        return self.constraints.check_property(lambda c: isinstance(c, (Counterbalance, AbsoluteRank)))
   
    @property
    def is_empty(self):
        # FIXME: not handled for replicated trials
        return not self.variables 

    @property
    def is_modified(self):
        return self.snapshot() != self.previous_snapshot
    
    @property
    def is_random(self):
        return not self.counterbalanced and not self.is_empty
    

    def num_plans(self):
        self._determine_num_plans()
        return 1 if self.is_empty or self.is_random else self._determine_num_plans()
    
    def to_latex(self):
        # FIXME: won't work with random plans
        matrix = self.get_plans()
        formatter = LatexExport(matrix)
        formatter.to_latex()

    def num_trials(self, n):
        self.trials = n
        return self
    
    def between_subjects(self, variable):
        self.add_variable(variable)
        self.trials = 1 if not self.trials else self.trials

        # enforces repeating trials when specified with within subjects variables
        self.add_constraint(
            InnerBlock(
                variable,
                0,
                1,
                [1,1]
                )
            )
        return self
    
    def counterbalance(self, variable:ExperimentVariable, w = 0, h = 0, stride = [1, 1]):
        self.add_constraint(Counterbalance(variable, width = w, height = h, stride = stride))
        return self
    
    def start_with(self, variable:ExperimentVariable, condition):
        condition = as_list(condition)
        rank = 1
        for c in condition:
            self.absolute_rank(variable, c, rank)
        rank+=1
  
        return self
    
    def set_position(self, variable:ExperimentVariable, condition, pos):
        self.add_constraint(SetPosition(variable, condition, pos))
        return self
    
    def set_rank(self, variable:ExperimentVariable, condition, rank, condition2):
        self.add_constraint(SetRank(variable, condition, rank, condition2))
        return self
    
    def absolute_rank(self, variable:ExperimentVariable, condition, rank):
        constraint = self.constraints.add_absolute_rank(variable, condition, rank)
        self.design_variables[variable].add_constraint(constraint)
        return self
    
    def add_constraint(self, constraint):
        self.constraints.add_constraint(constraint)
        self._add_design_variable(constraint.variable)
        self.design_variables[constraint.variable].add_constraint(constraint)

    def add_constraints(self, constraints:list):
        for c in constraints:
            self.add_constraint(c)

    def within_subjects(self, variable):
        self.add_variable(variable)
        self.trials = len(variable)
        self.add_constraint(NoRepeat(variable, width=self.trials))
        return self
    
    def limit_groups(self, n):
        self.groups.set_num_plans(n)
        return self

    def snapshot(self):
        groups = self.num_plans()
        constraint_ids = self.constraints.stringified()
        signature = "_".join(constraint_ids) + f"_{self.get_width()}_{groups}"
        return hashlib.sha256(signature.encode()).hexdigest()
    
    def get_width(self):
        return self.trials 
    
    def extract_counterbalance_info(self, var):
        """Extract variables and condition count"""
        return (len(var.get_variables()), len(var))
    
    def calculate_num_plans(self, counterbalanced_groups, rankings, num_trials):
        """Determine the number of experimental plans based on constraints and trial width."""
        total_n_plans = 1

        for variables, num_conditions in counterbalanced_groups:
                num_trials = num_trials
                total_n_plans *= calculate_plan_multiplier(num_conditions, variables, num_trials)
        for ranking in rankings:
            total_n_plans *= factorial_product_of_counts(ranking)

        return int(total_n_plans)
    
    def _determine_num_plans(self):
        """Determine the number of experimental plans based on constraints and trial width."""
        if len(self.groups) > 0:
            return len(self.groups)  # already specified by user

        counterbalance_info = []
        rankings = []

        for variable in self.design_variables:
            if self.design_variables[variable].is_counterbalanced:
                group = self.extract_counterbalance_info(variable)
                counterbalance_info.append(group)

            elif self.design_variables[variable].is_ranked:
                rankings.append(count_values(self.design_variables[variable].get_ranks()))

        return self.calculate_num_plans(counterbalance_info, rankings, self.trials)

    def test_eval(self):
        self.designer.start(self)
        return self.designer.eval_all()
        

    def identify_random_vars(self):
        return [
            v for v, obj in self.design_variables.items()
            if not (obj.is_counterbalanced or obj.is_ranked)
        ]
            
    def _add_design_variable(self, variable):
        if variable not in self.design_variables:
            self.design_variables[variable] = DesignVariable(variable)

    def add_variable(self, variable):
        assert isinstance(variable, ExperimentVariable)

        # FIXME
        if variable in self.design_variables: 
            raise ValueError(f"Cannot add variable '{variable}' — it already exists in design.")
        
        self._add_design_variable(variable)

        if isinstance(variable, MultiFactVariable):
            variables = variable.variables
        else:
            variables = [variable]
    
        self.variables.extend(variables)
        # ensures no duplicates and preserves order 
        self.variables = list(dict.fromkeys(self.variables))

    def add_variables(self, variables:list):
        for v in variables:
            self.add_variable(v)

    def get_constraints(self):
        return self.constraints.get_constraints()
    



class DesignVariable:
    def __init__(self, variable):
        self.variable = variable
        self.constraint_spec = {
            "Counterbalance":None,
            "NoRepeat":None,
            "AbsoluteRank":None,
            "OuterBlock":None,
            "InnerBlock":None,
        }


    def add_constraint(self, constraint):
        if self.constraint_spec[constraint.__class__.__name__] is None:
            self.constraint_spec[constraint.__class__.__name__] = constraint

    def has(self, constraint_name):
        return self.constraint_spec.get(constraint_name) is not None
    
    def get_ranks(self):
                # count_values(constraint.ranks)
        return self.constraint_spec.get("AbsoluteRank").ranks

    @property
    def is_counterbalanced(self):
        return self.has("Counterbalance")

    @property
    def is_repeated(self):
        return not self.has("NoRepeat")

    @property
    def is_ranked(self):
        return self.has("AbsoluteRank")

    @property
    def is_blocked_outer(self):
        return self.has("OuterBlock")

    @property
    def is_blocked_inner(self):
        return self.has("InnerBlock")

    @property
    def is_random(self):
        return not (
            self.is_counterbalanced
            or self.is_ranked
        )
    
    def get_width(self, width):
        div = 1

        outer_block = self.constraint_spec["OuterBlock"]
        if outer_block is not None:
            width = outer_block.width
        if self.is_blocked_inner:
            div = self.constraint_spec["InnerBlock"].width
  
        return int(width/div)
    

    def get_span(self):
        span = 1
        if self.is_blocked_inner:
            span = self.constraint_spec["InnerBlock"].width
               
        return span
        


from z3 import *
from .blocks import BlockFactor
import duckdb
import time
from planet.candl import *

class Unit:
    """ a Unit represents a singular unit that participates in a study. Units
            get randomly assigned to an experimental group during assignment

    n: id of the unit
    """
    def __init__(self, n):
        self.n = n

class Units:
    def __init__(self, n, table = "participants"):
        self.n = n
        self.attributes = []
        self.hash = str(time.time()).replace(".", "")
        self.table = table + self.hash
        self.evaled = False

    def add_attribute(self, attr):

        assert isinstance(attr, BlockFactor)
        self.attributes.append(attr)

    def _create_new_table(self):
        if self.evaled:
            duckdb.sql(f"TRUNCATE TABLE {self.table}")
        else:
            duckdb.sql(f"create table {self.table} ( pid int, plan int )")

    # FIXME: ugly workaround when assigning units multiple times
    def eval(self):
        self._create_new_table()
        duckdb.sql(f"INSERT INTO {self.table} SELECT i + 1, 0 FROM range({self.n}) AS t(i)")
        self.evaled = True


    def get_table(self):
        self.eval()
        return self.table

    def get_plan(self, n):
        return duckdb.execute(f"select plan from {self.table} where pid = {n}").fetchone()[0]

    def get_attributes(self):
        return self.attributes
    
    def __len__(self):
        return self.n 
    

class Clusters(Units):
    """A Cluster represents a group of Units that are grouped together for a specific purpose."""
    def __init__(self, units, n):
        assert units.n % n == 0
        super().__init__(n, "clusters")

        assert units is not None
        self.units = units

    def get_units(self, n):
        ret = duckdb.sql(f"select subunits from {self.table} where pid = {n}").fetchnumpy()["subunits"][0]
        return list(ret)

    def eval(self):
        self.units.eval()
        super().eval()

        duckdb.sql(f"""ALTER TABLE {self.table}
                        ADD subunits INTEGER[]""")
        
     
     
        limit = int(self.units.n/self.n)
        for i in range(1, self.n+1):   
            
            duckdb.sql(f"""update {self.table} set subunits = (select array_agg(units.pid) from 
                                    (select      *, 
                                                row_number() over(order by uuid()) rand
                                    from        {self.units.table} where pid not in (select unnest(subunits) from {self.table} ) limit {limit}) units)
                                    
                                    where pid = {i}""")
            



       

# inherit from variable? 
class Participants(Units):
    """ a Unit represents a singular unit that participates in a study. Units
            get randomly assigned to an experimental group during assignment

    n: id of the unit
    """
    def __init__(self, n=0):
        Units.__init__(self, n)

   

# inherit from variable? 
class Groups(Units):
    """ a Unit represents a singular unit that participates in a study. Units
            get randomly assigned to an experimental group during assignment

    n: id of the unit
    """
    def __init__(self, n=0):
        self.n = n

    def set_num_plans(self, n):
        self.n = n 
    
    def __len__(self):
        return self.n
    

 
    




        
    



from z3 import *
import numpy as np
from .candl import *
from .helpers import *
from .bitvector import BitVectors
import itertools
import warnings
from itertools import combinations

class Solver: 
    def __init__(self, shape, variables):
        
        # how can we use translate without circular dependencies? 
        # some of these vars are related and should be stored together as a struct 
        self.shape = shape
        self.variables = variables
        self.solver = z3.Optimize()

class BitVecSolver(Solver):
    def __init__(self, shape, variables):
        super().__init__(shape, variables)
        self.bitvectors = BitVectors(np.prod(self.shape), self.variables)
        # NOTE: z3 vectors is stored as a flattened array
        self.z3_variables = self.bitvectors.get_variables()
        self.constrain_z3_values()
        self.distinguish_rows()

    # you can come up with a better name
    def constrain_z3_values(self):
        """ for all of the z3 variables relating to a specific variable, 
         ensure that it can only be assigned to one of the levels
         of the specific variable """
        for variable in self.variables:
            for index in range(np.prod(self.shape)):
                lo = 0
                hi = len(variable) - 1
                masked_bits = self.bitvectors.get_variable_assignment(variable, self.z3_variables[index])
                self.solver.add(
                        And(
                            lo <= masked_bits, 
                            hi >= masked_bits
                        )  
                )
  
    def get_partition(self, width, stride):
        dim_variables = shape_array(self.z3_variables, self.shape)
        if width is not None:
            dim_variables = partition_matrix_by_columns(dim_variables, width, stride)
        return dim_variables

    def all_different(self, variables=None, width = None, stride = 1):
        matrix = self.get_partition(width, stride)
        # This is necessary when the combinations of variables are unique, but
        # not the individual variables
        variable_assignment = lambda row: [self.bitvectors.get_variable_assignment(v, row) for v in variables]
        for row in matrix:
            self.solver.add(at_least_one_diff(list(map(variable_assignment, row))))
        
    def n_trials(self):
        return self.shape[1]

    # NOTE: no rows can repeat in a given matrix 
    def distinguish_rows(self):
        plans = shape_array(self.z3_variables, self.shape)
        for i, j in combinations(range(len(plans)), 2):
            # At least one assignment must differ
            self.solver.add(Or([a != b for a, b in zip(plans[i], plans[j])]))



    def slice_matrix(self, arr, block = []):
        def get_spec(block_data):
            return block_data[0], block_data[1], block_data[2]
        start_column, end_column, column_step = get_spec(block[0])
        start_row, end_row, row_step = get_spec(block[1])
        return np.array(arr)[start_column:end_column:column_step, start_row:end_row:row_step]


    def absolute_rank(self, var, ranks):
        design_matrix = shape_array(self.z3_variables, self.shape)
        n = self.n_trials()

        rank_fn = Function(f'rank_{str(var)}', BitVecSort(self.bitvectors.determine_num_bits()), IntSort())
        self.solver.add([rank_fn(plan[i]) >= rank_fn(plan[i+1]) for i in range(n-1) for plan in design_matrix])

        # for use in quantified expression
        x = self.bitvectors.new_bitvector()
        for condition, rank in ranks.items():
            # set the rank for every condition of the respective variable
            self.solver.add(
                ForAll([x],
                    Implies(
                        self.bitvectors.get_variable_assignment(var, x) == condition,
                        rank_fn(x) == rank
                    )
                )
            )
            
    def count(self, variables, condition, f):
        """
        if f returns true increase count by one for all variables
        """
        counts = [If(f(var, condition), 1, 0) for var in variables]
        return sum(counts)

    def counterbalance(self, block=[], variables=None):
        """
        Apply counterbalancing to ensure equal occurrence of variable combinations.
            
        Returns:
            None: Modifies the solver in-place
        """
        design_matrix = shape_array(self.z3_variables, self.shape)
        design_matrix = self.slice_matrix(design_matrix, block)
        design_matrix = np.transpose(design_matrix)
      
        # Generate all possible combinations of variable values
        possible_conditions = list(itertools.product(
            *[set(range(len(variable))) for variable in variables]
        ))
        
        counts = []
        for trials in design_matrix:
            masked_variables = list(zip(*[
                list(map(lambda z3_variable: self.bitvectors.get_variable_assignment(variable, z3_variable), trials)) 
                for variable in variables
            ]))
            
            check_equality = lambda variables, assignments: And([variables[i] == assignments[i] for i in range(len(variables))])
            counts.extend([self.count(masked_variables, condition, check_equality) for condition in possible_conditions])
        
        # Add constraints to ensure equal counts for all combinations
        for i in range(len(counts)):
            for j in range(i+1, len(counts)):
                self.solver.add(counts[i] == counts[j])

   
    def match_block(self, variable, block = []):
        plans = shape_array(self.z3_variables, self.shape)
        block = self.slice_matrix(plans, block)

        flat_block = np.array(block).flatten()
        for i in range(len(flat_block)):
            for j in range(i, len(flat_block)):
                a1 = self.bitvectors.get_variable_assignment(variable, flat_block[i])
                a2 = self.bitvectors.get_variable_assignment(variable, flat_block[j])
                self.solver.add(a1==a2)

        # works when z3 representation is a matrix 
    def get_one_model(self):
        all_assignments = []
    
        if (self.solver.check() == sat):
            all_assignments = []
            model = self.solver.model()
      
            for var in self.z3_variables:
                all_assignments.append(model.evaluate(model[var]))
 
        return all_assignments
    
    # works when z3 rep iterates through many arrays
    # NOTE: should I merge these two functions?
    def get_all_models(self):
        all_orders = []
        self.solver.push()
        count = 0
        capacity = 10000
        while self.solver.check() == sat and count < capacity:
            model = self.solver.model()
            block = []
            order = []
            for var in self.z3_variables:
                block.append(var != model[var])
                order.append(model.evaluate(model[var]))

            all_orders.append(order)
            self.solver.add(Or(block))
            count+=1
        if self.solver.check() == sat:
            warnings.warn(f"There are more orders not captured in the model due order capacity (currently set to {capacity})")

        self.solver.pop()
        return all_orders


    def encoding_to_name(self, all_orders, variables):
        # this makes life much easier
        z3_assignments = flatten_array(all_orders)
        num_vars = len(variables)
        
        orders = []
        for z3_assignment in z3_assignments:
            # this is an integer value, represented as a binary string
            assignment = int(str(z3_assignment))
            # this is where we will store the string rep of a condition
            # which is a concatination of it's variable assignments
            decoded_assignment = ""
 
            # # there is probably a more efficient way to do this
            for i in range(num_vars):
                # get the ith variable's assignment
                variable_assignment = assignment 
                # shift right variable assignment unitl desired var assignment 
                # is in the lower bits (+1 to account for two's compliment)
         
                for j in range(i):
                    variable_assignment >>= get_num_bits(len(variables[j])) + 1

                # get lower bits (associated with given variable)
                mask =  all_ones(get_num_bits(len(variables[i]))+ 1)
                variable_assignment &= mask

                # FIXME
                if variable_assignment >= len(variables[i]):
                    condition = "dropout"
                else:
              
                    # FIXME: this is silly, but need to inverse order to match array indexing
                    # variable_assignment = len(variables[i]) - 1 - variable_assignment
                    condition = variables[i].conditions[variable_assignment]

                # conder variable assignment to a string representation 
                decoded_assignment+=str(condition)
                # FIXME; hyphenate output for readability
                if i < len(variables)-1:
                    decoded_assignment+="-"

            orders.append(decoded_assignment)

        # return back in original form
        return shape_array(orders, np.array(all_orders).shape)
    
    def name_to_encoding(self, model):

        # this makes life much easier
        model_assignments = flatten_array(model)
        num_vars = len(self.variables)
        
        for i in range(len(model_assignments)):
            assignment = model_assignments[i].split("-")

            # there is probably a more efficient way to do this
            for j in range(num_vars):
                # get the ith variable's assignment
                variable_assignment = assignment[j]
                condition = self.variables[j].condition_map[variable_assignment]
                __z3__ = self.bitvectors.get_variable_assignment(self.variables[j], self.z3_variables[i])
                self.solver.add(__z3__ == condition)

    def check_model(self, model):
  
        self.solver.push()

        self.name_to_encoding(model, self.variables)
        result = self.solver.check()

        self.solver.pop()

        return result
